/*
 *  armadillo2x0_sw.c
 *
 *  Copyright (C) 2006 Atmark Techno, Inc.
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>

#include "armadillo2x0_sw.h"

#ifndef SW_MINOR
#define SW_MINOR 187
#endif
#define SW_NAME "sw"

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-2x0 Takt Switch driver"
#define LICENSE     "GPL v2"

#define VERSION     SW_NAME ": " DESCRIPTION ", (C) 2006 " AUTHOR

static DECLARE_WAIT_QUEUE_HEAD(sw_wait);
static volatile int reset = 1;

static ssize_t armadillo2x0_sw_read (struct file *file, char __user *buf,
				      size_t count, loff_t *ppos)
{
	char val = inl(GPIO_PFDR) >> 7;

	reset = 1;

	if (copy_to_user(buf, &val, 1))
		return -EFAULT;

        return 1;
}

static int armadillo2x0_sw_open(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev); 
	if (minor != SW_MINOR)
		return -ENODEV;

#ifdef MODULE
	//MOD_INC_USE_COUNT;
#endif

	return 0;
}

static int armadillo2x0_sw_release(struct inode *inode, struct file *file)
{

#ifdef MODULE
	//MOD_DEC_USE_COUNT;
#endif

	return 0;
}

static irqreturn_t armadillo2x0_sw_interrupt_ack
                   (int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long ulTemp;

	if (!(inl(GPIO_INTSTATUSF) & 0x80)) {
		return IRQ_NONE;
	}

	outl(0x80, GPIO_FEOI);
	ulTemp = inl(GPIO_FEOI);

	reset = 0;
	if (waitqueue_active(&sw_wait)) {
		wake_up_interruptible(&sw_wait);
	}

	return IRQ_HANDLED;
}

static int armadillo2x0_sw_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	static int irq_disabled = 1;

	switch(cmd) {
		case ARMADILLO2X0_SW_REQUEST_IRQ:
			if (!irq_disabled) {
			  return 0;
			}
			irq_disabled = request_irq
			  (IRQ_GPIO7, armadillo2x0_sw_interrupt_ack,
			   SA_INTERRUPT, SW_NAME, NULL);
			if (irq_disabled) {
				return -EFAULT;
			}
			break;

		case ARMADILLO2X0_SW_FREE_IRQ:
			if (!irq_disabled) {
				free_irq(IRQ_GPIO7, NULL);
				irq_disabled = 1;
			}
			break;

		default:
			return -ENOIOCTLCMD;
	}

	return 0;
}

static unsigned int armadillo2x0_sw_poll(struct file *file, poll_table *wait)
{
	poll_wait(file, &sw_wait, wait);
	if (!reset) {
		return POLLIN | POLLRDNORM;
	}

	return 0;
}

static struct file_operations armadillo2x0_sw_fops =
{
	owner:		THIS_MODULE,
	read:		armadillo2x0_sw_read,
	poll:		armadillo2x0_sw_poll,
	ioctl:		armadillo2x0_sw_ioctl,
	open:		armadillo2x0_sw_open,
	release:	armadillo2x0_sw_release,
};

static struct miscdevice armadillo2x0_sw_miscdev =
{
	SW_MINOR,
	SW_NAME,
	&armadillo2x0_sw_fops
};

int __init armadillo2x0_sw_init(void)
{
	outl(inl(GPIO_PFDDR) & ~0x80, GPIO_PFDDR);
	outl(inl(GPIO_FINTTYPE1) | 0x80, GPIO_FINTTYPE1);
	outl(inl(GPIO_FINTTYPE2) & ~0x80, GPIO_FINTTYPE2);
	outl(inl(GPIO_FDB) | 0x80, GPIO_FDB);
	outl(inl(GPIO_FINTEN) | 0x80, GPIO_FINTEN);

	misc_register(&armadillo2x0_sw_miscdev);

	printk(KERN_INFO VERSION "\n");

	return 0;
}	

void __exit armadillo2x0_sw_exit(void)
{
	misc_deregister(&armadillo2x0_sw_miscdev);

	outl(inl(GPIO_FINTEN) & ~0x80, GPIO_FINTEN);
}

module_init(armadillo2x0_sw_init);
module_exit(armadillo2x0_sw_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
